//***************************************************************************
// Group Linslot / Linux - Slotrace Manager
// File arduino.cc
// Date 16.11.08 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//***************************************************************************

//***************************************************************************
// Include
//***************************************************************************

#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>

#include <arduino.hpp>

//***************************************************************************
// 
//***************************************************************************

const char* cmdGetTime = "GETTIME;";

//***************************************************************************
// Object
//***************************************************************************

Arduino::Arduino()
   : IoInterface()
{
   fdDevice = 0;
   readTimeout = 60;
   writeTimeout = 60;

   tvNow(&tvBoarStartTime);
   tvNow(&tvLastIoTime);

   bzero(&oldtio, sizeof(oldtio));

   out[0] = 0xFF;
   out[1] = 0xFF;
}

Arduino::~Arduino()
{
   close();
}

//***************************************************************************
// 
//***************************************************************************

int Arduino::open(const char* dev)
{
   struct termios newtio;

   if (!dev)
      dev = deviceName;
   else
      strcpy(deviceName, dev);

   if (!deviceName || !*deviceName)
      return fail;

   QMutexLocker locker(&deviceMutex);

   if ((fdDevice = ::open(deviceName, O_RDWR | O_NOCTTY)) < 0)
   {
      tell(eloAlways, "Error: Opening '%s' failed", deviceName);
      return fail;
   }

   // configure serial line

   tcgetattr(fdDevice, &oldtio);
   bzero(&newtio, sizeof(newtio));
   newtio.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
   newtio.c_iflag = IGNPAR | PARMRK;
   newtio.c_oflag = 0;
   newtio.c_lflag = 0;
   tcflush(fdDevice, TCIFLUSH);
   tcsetattr(fdDevice, TCSANOW, &newtio);

   // first ask about the boardtime

   write(fdDevice, cmdGetTime, strlen(cmdGetTime));

   return success;
}

int Arduino::close()
{
   // restore the old settings and close

   if (isOpen())
   {
      tell(eloDebug, "Debug: Closing connection");

      QMutexLocker locker(&deviceMutex);

      tcsetattr(fdDevice, TCSANOW, &oldtio);
      ::close(fdDevice);
      fdDevice = 0;
   }

   return success;
}

int Arduino::reopen(const char* dev)
{
   close();
   
   return open(dev);
}

int Arduino::flush()
{
   return success;
}

int Arduino::writeBit(int bit, int value)
{
   char cmd[sizeCmdMax+TB];
   unsigned char* byte = bit < 8 ? &out[0] : &out[1];

   setBitTo(*byte, bit < 8 ? bit : bit-8, value);

   tell(eloDebug, "Debug: Write bit (%d) value (%d)", bit, value);

   sprintf(cmd, "DO%d=%d;", bit, value);

   QMutexLocker locker(&deviceMutex);
   write(fdDevice, cmd, strlen(cmd));

   return success;
}

int Arduino::readOutBit(int bit)
{
   char buf[20+TB];
   unsigned char byte = bit < 8 ? out[0] : out[1];
   
   tell(eloDebug2, "Debug: Byte is '%s' on port %d", toBinStr(byte, buf), bit < 8 ? 0 : 1);

   return isBit(byte, bit < 8 ? bit : bit-8);
}

int Arduino::writeByte(unsigned char /*byte*/, int /*byteNum*/)
{
   // not needed ... ?

   return success;
}

int Arduino::readByte(unsigned char* byte, int byteNum)
{
   char b;
   char cmd[sizeCmdMax+TB];
   IoValue value;

   if (!isOpen())
      return fail;

   deviceMutex.lock();

   if (::read(fdDevice, &b, 1) <= 0)
   {
      // no input data pending

      deviceMutex.unlock();
      return fail;
   }

   tell(eloDebug2, "Debug: Got lookahead byte '0x%X'", b);
   deviceMutex.unlock();
   readCmd(cmd, b);

   if (dispatch(cmd, value) != cmdDigitalInput)
      return ignore;

   if (byte)
      *byte = value.bytes[byteNum];

   return success;
}

int Arduino::lookByte(int /*byteNum*/, unsigned  char* /*byte*/)
{
   // not needed ... ?

   return success;
}

int Arduino::setTimeout(unsigned int timeout)
{
   readTimeout = timeout / 1000;
   return success;
}

int Arduino::setWriteTimeout(unsigned int timeout)
{
   writeTimeout = timeout / 1000;
   return success;
}

//***************************************************************************
// Read Command
//***************************************************************************

const char* Arduino::readCmd(char* command, char lookahead)
{
   int count = 0;
   char b;
   int start = time(0);
   int cnt;

   *command = 0;

   if (!isOpen())
      return command;

   if (lookahead && lookahead != '\n')
   {
      command[0] = lookahead;
      command[1] = 0;
      count++;
   }

   // QMutexLocker locker(&deviceMutex);

   while (time(0) < start + readTimeout)
   { 
      deviceMutex.lock();
      cnt = ::read(fdDevice, &b, 1);
      deviceMutex.unlock();

      if (cnt <= 0)
      {
         usleep(100);
         continue;
      }

      // skip cartridge return 

      if (b == '\r')
         continue;

      // stop at linefeed

      if (b == '\n')
         break;

      command[count] = b;
      count++;      
   }

   command[count] = 0;

   tell(eloDebug2, "Debug: <- (%d)[%s]", count, command);

   return command;
}

int Arduino::dispatch(char* buf, IoValue& value)
{
   // Examples:
   //     "BOARDTIME=750205;"
   //     "DI:754041=2999;"

   char* p = buf;
   char* cmd = buf;
   char* val = 0;
   char* t = 0;

   if ((p = strchr(p, ':')))
   {
      *p = 0;
      p++;
      t = p;
   }
   else
      p = buf;

   if (!(p = strchr(p, '=')))
      return fail;

   *p = 0;
   p++;
   val = p;

   if (!(p = strchr(p, ';')))
      return fail;

   *p = 0;

   value.dword = atoi(val);

   if (t) 
      tvLastIoTime = addMs2Tv(tvBoarStartTime, atol(t));

   char tmp[20+TB];

   tell(eloDebug2, "Debug: Got command '%s' - '%s' at (%sms)",
        notNull(cmd), toBinStr(atoi(val), tmp), notNull(t));

   if (strcasecmp(cmd, "BOARDTIME") == 0)
   {
      tvNow(&tvBoarStartTime);
      tvBoarStartTime = addMs2Tv(tvBoarStartTime, atol(val));
      return cmdBoardTime;
   }

   if (strcasecmp(cmd, "DI") == 0)
      return cmdDigitalInput;

   if (strcasecmp(cmd, "AI") == 0)
      return cmdAnalogInput;

   return fail;
}
