//***************************************************************************
// Group Linslot / Linux - Slotrace Manager
// File  main.cpp
// Date 26.01.08 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//--------------------------------------------------------------------------
// Programm Part for Arduino-Diecimila
//***************************************************************************

//***************************************************************************
// Includes
//***************************************************************************

#include <WProgram.h>

#define TIMER_CLOCK_FREQ 2000000.0 // 2MHz for /8 prescale from 16MHz

//**************************************************************
// Definitions
//**************************************************************

enum Misc
{
  fail        = -1,
  success     = 0,
  na          = -1,
  TB          = 1,

  maxCmd      = 100,
  cmdEnd      = ';',
  prellTime   = 200,

  firstInBit       = 2,
  lastInBit        = 3,
  pmwOut1          = 5,
  pmwOut2          = 6,

  firstExtendedOut = 14,
  lastExtendedOut  = firstExtendedOut + 7,

  bitShiftRegSlave   = 10,
  bitShiftRegDataOut = 11,
  bitShiftRegDataIn  = 12,
  bitShiftRegClock   = 13
};

//**************************************************************
// Globals
//**************************************************************

volatile unsigned long ioTimes[14];
volatile unsigned int lastInputValue = 0xFF;
volatile unsigned long ioTime = 0;
volatile int ioPending;
volatile int timerLoadValue = 0;
volatile int irqScale = 0;
volatile int ghostcarPin = na;
volatile char shiftValue = 0;

//***************************************************************************
// Prototypes
//***************************************************************************

void sendAnalogValue(byte pin);
void ioInterrupt();
unsigned char setupTimer2(float irqFreq);

//***************************************************************************
// Debug
//***************************************************************************

void debug(const char* msg)
{
  Serial.print("DEBUG ");
  Serial.print(msg);
  Serial.println(";");
}

//**************************************************************
// Setup
//**************************************************************

void setup()
{
   byte clr;

   // serial interface

   Serial.begin(9600);

   // digital io pins

   for (int i = firstInBit; i <= lastInBit; i++)
      pinMode(i, INPUT);

   pinMode(pmwOut1, OUTPUT);
   pinMode(pmwOut2, OUTPUT);

   // init shift register control pins

   pinMode(bitShiftRegClock, OUTPUT);
   // digitalWrite(bitShiftRegClock, LOW);
   pinMode(bitShiftRegDataOut, OUTPUT);
   // digitalWrite(bitShiftRegDataOut, LOW);

   pinMode(bitShiftRegDataIn, INPUT);

   pinMode(bitShiftRegSlave, OUTPUT);
   digitalWrite(bitShiftRegSlave, HIGH);  // disable device

   // Now we set the SPI Control register (SPCR) to the binary value 01010000. 
   // In the control register each bit sets a different functionality. The eighth 
   // bit disables the SPI interrupt, the seventh bit enables the SPI, the sixth 
   // bit chooses transmission with the most significant bit going first, the fifth
   // bit puts the Arduino in Master mode, the fourth bit sets the data clock idle 
   // when it is low, the third bit sets the SPI to sample data on the rising edge
   // of the data clock, and the second and first bits set the speed of the SPI to 
   // system speed / 4 (the fastest). After setting our control register up we read 
   // the SPI status register (SPSR) and data register (SPDR) in to the junk clr 
   // variable to clear out any spurious data from past runs:

   SPCR = (1 << SPE) | (1 << MSTR);

   // SPI als Master, High-Bits zuerst, SCK ist HIGH wenn inaktiv 
   // SPCR = (1 << SPE) | (1 << MSTR) | (1 << CPOL);
   // maximale Geschwindigkeit: F_CPU / 2 
   // SPSR |= (1 << SPI2X);

   clr = SPSR;
   clr = SPDR;
   delay(10);

   // init interrupt stuff

   attachInterrupt(0, ioInterrupt, CHANGE); 
   timerLoadValue = setupTimer2(10000);  // set interrupt to 10kHz

   ioPending = false;
}

//***************************************************************************
// Setup Timer 2
//    Configures the 8 Bit Timer2 to generate an interrupt
//    at the specified frequency.
//    Returns the timer load value which must be loaded into TCNT2
//    inside your ISR routine.
//***************************************************************************

unsigned char setupTimer2(float irqFreq)
{
   unsigned char result; // The timer load value.

   // Calculate the timer load value
   // The 257 really should be 256 but I get better results with 257.

   result = (int)((257.0-(TIMER_CLOCK_FREQ/irqFreq))+0.5);

   // Timer2 Settings: Timer Prescaler /8, mode 0
   // Timer clock = 16MHz/8 = 2Mhz or 0.5us
   // The /8 prescale gives us a good range to work with
   // so we just hard code this for now.

   TCCR2A = 0;
   TCCR2B = 0<<CS22 | 1<<CS21 | 0<<CS20;

   TIMSK2 = 1<<TOIE2;  // Timer2 Overflow Interrupt Enable
   TCNT2 = result;     // load the timer for its first cycle

   return result;
}

//**************************************************************
// IRQ Routines
//**************************************************************

void ioInterrupt()
{
  ioPending = true;
  ioTime = millis();
}

//***************************************************************************
// Timer2 overflow interrupt vector handler
//***************************************************************************

ISR(TIMER2_OVF_vect) 
{
   if (!irqScale--)
   {
      // Toggle the IO pin to the other state.
      
      digitalWrite(7, !digitalRead(7));

      if (ghostcarPin != na)
         sendAnalogValue(ghostcarPin);

      irqScale = 10;
   }

   // Capture the current timer value. This is how much error we
   // have due to interrupt latency and the work in this function

   int latency = TCNT2;

   // Reload the timer and correct for latency.

   TCNT2 = latency + timerLoadValue;
}

//**************************************************************
// Write digital out via shift register
//**************************************************************

void shiftRegWrite(byte pin, byte value)
{
   int setMask = 1 << pin;

   if (value && !(shiftValue & setMask))
   {
      // set requested and not already set -> set now ;)
      
      shiftValue |= setMask;
   }

   else if (!value && shiftValue & setMask)
   {
      // clear requested and not already cleared -> clear now ;)

      shiftValue &= ~setMask;
   }

   else
      return ;

   // bit changed, we have to shift all bits
   
   digitalWrite(bitShiftRegSlave, LOW);  // enable device

   SPDR = shiftValue;              // start the transmission

   while (!(SPSR & (1 << SPIF)))     // wait the end of the transmission
      ;

   digitalWrite(bitShiftRegSlave, HIGH);  // disable device

//    for (int i = 7; i >= 0; i--)
//    {
//       // data

//       if (shiftValue & (1 << i))
//          digitalWrite(bitShiftRegData, HIGH);
//       else
//          digitalWrite(bitShiftRegData, LOW);

//       // clock impulse (duration due do digital write about 5s)

//       digitalWrite(bitShiftRegClock, HIGH);
//       digitalWrite(bitShiftRegClock, LOW);
//    }

   // latch impulse
   
//    digitalWrite(bitShiftRegLatch, HIGH);
//    digitalWrite(bitShiftRegLatch, LOW);
}

//**************************************************************
// 
//**************************************************************

void sendAnalogValue(byte pin)
{
   Serial.print("AI:");
   Serial.print(pin, DEC);
   Serial.print("=");
   Serial.print(analogRead(pin));
   Serial.println(";");
}

//**************************************************************
// 
//**************************************************************

void sendDigitalIoState()
{ 
  int b, mask;
  unsigned int value = lastInputValue;
  byte changes = 0;

  for (int bit = firstInBit; bit <= lastInBit; bit++)
  {
    b = digitalRead(bit) << bit;
    mask = 1 << bit;

//     if (b)
//     {
//       value |= mask;
//     }
//     else if ((lastInputValue ^ b) && (millis() - ioTimes[bit] > prellTime))

    if (((lastInputValue & mask) ^ b) && (millis() - ioTimes[bit] > prellTime))
    {
      // bit changed -and- last change older than 'prellTime'

      ioTimes[bit] = ioTime;

      if (b)
         value |= mask;
      else
         value &= ~mask;
      changes++;
    }
  }

  if (changes)
  {
    Serial.print("DI:");
    Serial.print(ioTime);
    Serial.print("=");
    Serial.print(value);
    Serial.println(";");
  }

  lastInputValue = value;    
  ioPending = false;
}

//**************************************************************
// Command GETTIME
//**************************************************************

void cmdGettime(const char* cmd)
{
  Serial.print("BOARDTIME=");
  Serial.print(millis());
  Serial.println(";");
}

//**************************************************************
// Command DO
//**************************************************************

void cmdDigitalOut(const char* cmd)
{
  char* p = strchr(cmd, '=');
  int pin;

  if (!p) return ;

  *p = 0;
  p++;

  pin = atoi(cmd);
  
  if (pin >= firstExtendedOut && pin <= lastExtendedOut)
     shiftRegWrite(pin - firstExtendedOut, atoi(p));
  else
     debug("Error: No output bit");
  // digitalWrite(pin, atoi(p));
}

//**************************************************************
// Command AO
//**************************************************************

void cmdAnalogOut(const char* cmd)
{
  char* p = strchr(cmd, '=');

  if (!p) return ;

  *p = 0;
  p++;

  analogWrite(atoi(cmd), atoi(p));
}

//**************************************************************
// Command RGC 'Record - Ghost Car'
//**************************************************************

void cmdRecordGhostCar(const char* cmd)
{
  unsigned int cycle;
  unsigned int pin;
  char* p;

  if (!(p = strchr(cmd, '=')))
    return ;

  p++;

  if ((cycle = atoi(p)) < 10)
    return ;
    
  if (!(p = strchr(cmd, ':')))
    return ;

  p++;

  if ((pin = atoi(p)) < 0 || pin > 5)
    return ;
    
  ioPending = false;
  ghostcarPin = pin;

  while (!ioPending && !Serial.available())
  {
     // sendAnalogValue(pin);
     // delay(cycle);
  }

  ghostcarPin = na;
  ioPending = false;

  debug("RGC ready!");
}

//**************************************************************
// Command Parser
//**************************************************************

int getCommand(char* cmd)
{
  int i = 0;
  int c;
  
  while (i < maxCmd)
  {
    if ((c = Serial.read()) > 0)
    {
      if (c == cmdEnd)
        break;

      cmd[i++] = c;
    }
  }

  cmd[i] = 0;
  
  // debug(cmd);
  
  return success;
}

//**************************************************************
// Main Loop
//**************************************************************

void loop()
{
  char cmd[maxCmd+TB];
  *cmd = 0;
  
  // analogWrite(3, 128);

  // Command Dispatcher

  if (Serial.available())
  {
    getCommand(cmd);

    if (strcasecmp(cmd, "GETTIME") == 0)
      cmdGettime(cmd);

    else if (strncasecmp(cmd, "DO", 2) == 0)
      cmdDigitalOut(cmd+2);

    else if (strncasecmp(cmd, "AO", 2) == 0)
      cmdAnalogOut(cmd+2);

    else if (strncasecmp(cmd, "RGC", 3) == 0)
      cmdRecordGhostCar(cmd);
  }

  // Digital IO

  if (ioPending)
  {
    sendDigitalIoState();
  }
}

//***************************************************************************
// Main
//***************************************************************************

int main()
{
	init();
	setup();

	for (;;)
		loop();

	return 0;
}
